@@ -222,6 +222,14 @@ gail_focus_watcher (GSignalInvocationHint *ihint,
         }
       else
         {
+          if (next_focus_widget)
+            {
+               GtkWidget *toplevel;
+
+               toplevel = gtk_widget_get_toplevel (next_focus_widget);
+               if (toplevel == widget)
+                 next_focus_widget = NULL; 
+            }
           /* focus out */
           widget = NULL;
         }
@@ -240,6 +248,14 @@ gail_focus_watcher (GSignalInvocationHint *ihint,
           return TRUE;
         }
     }
+  /*
+   * If the focus widget is a GtkSocket without a plug
+   * then ignore the focus notification as the embedded
+   * plug will report a focus notification.
+   */
+  if (GTK_IS_SOCKET (widget) &&
+      GTK_SOCKET (widget)->plug_widget == NULL)
+    return TRUE;
   /*
    * The widget may not yet be visible on the screen so we wait until it is.
    */
@@ -389,7 +405,8 @@ gail_deselect_watcher (GSignalInvocationHint *ihint,
         }
       else
         {
-          gail_focus_notify_when_idle (menu_shell);
+          if (!GTK_IS_MENU_BAR (menu_shell))
+            gail_focus_notify_when_idle (menu_shell);
         }
     }
   return TRUE; 
@@ -492,6 +509,13 @@ gail_focus_notify_when_idle (GtkWidget *widget)
     {
       if (widget)
         {
+          /*
+           * Ignore focus request when menu item is going to be focused.
+           * See bug #123432.
+           */
+          if (GTK_IS_MENU_ITEM (next_focus_widget) && !GTK_IS_MENU_ITEM (widget))
+            return;
+
           g_source_remove (focus_notify_handler);
           if (next_focus_widget)
             g_object_remove_weak_pointer (G_OBJECT (next_focus_widget), (gpointer *)&next_focus_widget);
