@@ -43,6 +43,9 @@ static void                  gail_button_init_textutil    (GailButton     *butto
 
 static void                  gail_button_pressed_enter_handler  (GtkWidget       *widget);
 static void                  gail_button_released_leave_handler (GtkWidget       *widget);
+static gint                  gail_button_real_add_gtk           (GtkContainer    *container,
+                                                                 GtkWidget       *widget,
+                                                                 gpointer        data);
 
 
 static void                  atk_action_interface_init  (AtkActionIface *iface);
@@ -207,7 +210,7 @@ gail_button_class_init (GailButtonClass *klass)
   class->ref_state_set = gail_button_ref_state_set;
   class->initialize = gail_button_real_initialize;
 
-  container_class->add_gtk = NULL;
+  container_class->add_gtk = gail_button_real_add_gtk;
   container_class->remove_gtk = NULL;
 }
 
@@ -360,6 +363,34 @@ gail_button_init_textutil (GailButton  *button,
                     button);     
 }
 
+static gint
+gail_button_real_add_gtk (GtkContainer *container,
+                          GtkWidget    *widget,
+                          gpointer     data)
+{
+  GtkLabel *label;
+  GailButton *button;
+
+  if (GTK_IS_LABEL (widget))
+    {
+      const gchar* label_text;
+
+      label = GTK_LABEL (widget);
+
+
+      button = GAIL_BUTTON (data);
+      if (!button->textutil)
+        gail_button_init_textutil (button, label);
+      else
+        {
+          label_text = gtk_label_get_text (label);
+          gail_text_util_text_setup (button->textutil, label_text);
+        }
+    }
+
+  return 1;
+}
+
 static void
 atk_action_interface_init (AtkActionIface *iface)
 {
