@@ -1,5 +1,5 @@
 /* GAIL - The GNOME Accessibility Enabling Library
- * Copyright 2001 Sun Microsystems Inc.
+ * Copyright 2001, 2002, 2003 Sun Microsystems Inc.
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -159,6 +159,7 @@ gail_label_real_initialize (AtkObject *obj,
                             gpointer  data)
 {
   GtkLabel  *label;
+  GtkWidget  *widget;
   GailLabel *gail_label;
   const gchar *label_text;
 
@@ -166,6 +167,8 @@ gail_label_real_initialize (AtkObject *obj,
   
   gail_label = GAIL_LABEL (obj);
 
+  gail_label->window_create_handler = 0;
+  gail_label->has_top_level = FALSE;
   gail_label->cursor_position = 0;
   gail_label->selection_bound = 0;
   
@@ -179,6 +182,26 @@ gail_label_real_initialize (AtkObject *obj,
     gail_label->label_length = 0;
   else
     gail_label->label_length = g_utf8_strlen (label_text, -1);
+
+  /* 
+   * Check whether ancestor of GtkLabel is a GtkButton and if so
+   * set accessible parent for GailLabel
+   */
+  widget = GTK_WIDGET (data);
+  while (widget != NULL)
+    {
+      widget = gtk_widget_get_parent (widget);
+      if (GTK_IS_BUTTON (widget))
+        {
+          atk_object_set_parent (obj, gtk_widget_get_accessible (widget));
+          break;
+        }
+    }
+
+  if (GTK_IS_ACCEL_LABEL (widget))
+    obj->role = ATK_ROLE_ACCEL_LABEL;
+  else
+    obj->role = ATK_ROLE_LABEL;
 }
 
 AtkObject* 
@@ -194,14 +217,61 @@ gail_label_new (GtkWidget *widget)
   accessible = ATK_OBJECT (object);
   atk_object_initialize (accessible, widget);
 
-  if (GTK_IS_ACCEL_LABEL (widget))
-    accessible->role = ATK_ROLE_ACCEL_LABEL;
-  else
-    accessible->role = ATK_ROLE_LABEL;
-
   return accessible;
 }
 
+static void
+notify_name_change (AtkObject *atk_obj)
+{
+  GtkLabel *label;
+  GailLabel *gail_label;
+  GtkWidget *widget;
+  GObject *gail_obj;
+  const gchar* label_text;
+
+  widget = GTK_ACCESSIBLE (atk_obj)->widget;
+  if (widget == NULL)
+    /*
+     * State is defunct
+     */
+    return;
+
+  gail_obj = G_OBJECT (atk_obj);
+  label = GTK_LABEL (widget);
+  gail_label = GAIL_LABEL (atk_obj);
+
+  /* Create a delete text and an insert text signal */
+ 
+  label_text = gtk_label_get_text (label);
+
+  g_signal_emit_by_name (gail_obj, "text_changed::delete", 0, 
+                         gail_label->label_length);
+  gail_text_util_text_setup (gail_label->textutil, label_text);
+  if (label_text == NULL)
+    gail_label->label_length = 0;
+  else
+    gail_label->label_length = g_utf8_strlen (label_text, -1);
+  g_signal_emit_by_name (gail_obj, "text_changed::insert", 0, 
+                         gail_label->label_length);
+
+  if (atk_obj->name == NULL)
+    /*
+     * The label has changed so notify a change in accessible-name
+     */
+    g_object_notify (gail_obj, "accessible-name");
+
+  g_signal_emit_by_name (gail_obj, "visible_data_changed");
+}
+
+static void
+window_created (GObject *obj,
+                gpointer data)
+{
+  g_return_if_fail (GAIL_LABEL (data));
+
+  notify_name_change (ATK_OBJECT (data)); 
+}
+
 static void
 gail_label_real_notify_gtk (GObject           *obj,
                             GParamSpec        *pspec)
@@ -211,32 +281,38 @@ gail_label_real_notify_gtk (GObject           *obj,
   GtkLabel *label;
   GailLabel *gail_label;
   GObject *gail_obj;
+  AtkObject *top_level;
+  AtkObject *temp_obj;
 
-  gail_obj = G_OBJECT (atk_obj);
-  label = GTK_LABEL (widget);
   gail_label = GAIL_LABEL (atk_obj);
 
   if (strcmp (pspec->name, "label") == 0)
     {
-      const gchar* label_text;
-
-      /* Create a delete text and an insert text signal */
- 
-      label_text = gtk_label_get_text (label);
-
-      g_signal_emit_by_name (gail_obj, "text_changed::delete", 0, 
-                             gail_label->label_length);
-      gail_text_util_text_setup (gail_label->textutil, label_text);
-      g_signal_emit_by_name (gail_obj, "text_changed::insert", 0, 
-                             gail_label->label_length);
-
-      if (atk_obj->name == NULL)
-        /*
-         * The label has changed so notify a change in accessible-name
-         */
-        g_object_notify (gail_obj, "accessible-name");
-
-      g_signal_emit_by_name (gail_obj, "visible_data_changed");
+     /* 
+      * We may get a label change for a label which is not attached to an
+      * application. We wait until the toplevel window is created before
+      * emitting the notification.
+      *
+      * This happens when [Ctrl+]Alt+Tab is pressed in metacity
+      */
+      if (!gail_label->has_top_level)
+        {
+          temp_obj = atk_obj;
+          while (temp_obj)
+            {
+              top_level = temp_obj;
+              temp_obj = atk_object_get_parent (top_level);
+            }
+          if (atk_object_get_role (top_level) != ATK_ROLE_APPLICATION)
+            {
+              if (gail_label->window_create_handler == 0)
+                gail_label->window_create_handler = g_signal_connect_after (top_level, "create", G_CALLBACK (window_created), atk_obj);
+            }
+          else
+            gail_label->has_top_level = TRUE;
+        }
+      if (gail_label->has_top_level)
+        notify_name_change (atk_obj);
     }
   else if (strcmp (pspec->name, "cursor-position") == 0)
     {
@@ -244,6 +320,9 @@ gail_label_real_notify_gtk (GObject           *obj,
       gboolean text_caret_moved = FALSE;
       gboolean selection_changed = FALSE;
 
+      gail_obj = G_OBJECT (atk_obj);
+      label = GTK_LABEL (widget);
+
       if (gtk_label_get_selection_bounds (label, &start, &end))
         {
           if (start != gail_label->cursor_position ||
@@ -335,7 +414,6 @@ gail_label_ref_relation_set (AtkObject *obj)
   if (widget == NULL)
     /*
      * State is defunct
-
      */
     return NULL;
 
